import json

from pytonconnect.provider import BridgeProvider as BaseBridgeProvider
from pytonconnect.storage import IStorage
from pytonconnect.logger import _LOGGER  # noqa


class BridgeProvider(BaseBridgeProvider):

    def __init__(self, storage: IStorage, wallet: dict, redirect_url: str = None) -> None:
        self.redirect_url = redirect_url
        super().__init__(storage, wallet)

    async def _gateway_listener(self, bridge_incoming_message):
        wallet_message = json.loads(
            self._session.session_crypto.decrypt(bridge_incoming_message['message'], bridge_incoming_message['from'])
        )

        _LOGGER.debug(f'Wallet message received: {wallet_message}')

        if 'event' not in wallet_message:
            if 'id' in wallet_message:
                id_ = wallet_message['id']
                if id_ not in self._pending_requests:
                    _LOGGER.debug(f"Response id {id_} doesn't match any request's id")
                    return

                self._pending_requests[id_].set_result(wallet_message)
                del self._pending_requests[id_]
            return

        if 'id' in wallet_message:
            id_ = int(wallet_message['id'])
            connection = json.loads(await self._storage.get_item(IStorage.KEY_CONNECTION, '{}'))
            last_id = connection['last_wallet_event_id'] if 'last_wallet_event_id' in connection else 0

            if last_id and id_ <= last_id:
                id_ = id_ - 1 if id_ == last_id else id_ - 2
                _LOGGER.error(
                    f'Received event id (={id_}) must be greater than stored last wallet event id (={last_id})')

            if 'event' in wallet_message and wallet_message['event'] != 'connect':
                connection['last_wallet_event_id'] = id_
                await self._storage.set_item(IStorage.KEY_CONNECTION, json.dumps(connection))

        # self.listeners might be modified in the event handler
        listeners = self._listeners.copy()

        if wallet_message['event'] == 'connect':
            await self._update_session(wallet_message, bridge_incoming_message['from'])

        elif wallet_message['event'] == 'disconnect':
            await self._remove_session()

        for listener in listeners:
            listener(wallet_message)

    def _generate_tg_universal_url(self, universal_url: str, request: dict):
        universal_url = universal_url.replace("attach=wallet", "startattach=")
        link = self._generate_regular_universal_url('about:blank', request)
        link_params = link.split('?')[1]

        link_params += f"&ret={self.redirect_url}" if self.redirect_url else "&ret=back"

        start_attach = (
                'tonconnect-' + link_params
                .replace('.', '%2E')
                .replace('-', '%2D')
                .replace('_', '%5F')
                .replace('&', '-')
                .replace('=', '__')
                .replace('%', '--')
                .replace('+', '')
                .replace(':', '--3A')
                .replace('/', '--2F')
        )
        return universal_url + start_attach



from typing import Callable, Dict, Any, Awaitable, Optional, Type, List, Union

from aiogram import BaseMiddleware
from aiogram.fsm.context import FSMContext
from aiogram.types import TelegramObject, User

from .manager import ATCManager
from .tonconnect import AiogramTonConnect
from .tonconnect.storage import SessionStorage
from .tonconnect.models import ATCUser, AppWallet, AccountWallet, InfoWallet
from .tonconnect.storage.base import ATCStorageBase
from .utils.keyboards import InlineKeyboardBase, InlineKeyboard
from .utils.qrcode import QRImageProviderBase, QRUrlProviderBase, QRImageProvider
from .utils.texts import TextMessageBase, TextMessage


class AiogramTonConnectMiddleware(BaseMiddleware):
    """
    Aiogram middleware for AiogramTonConnect integration.

    :param storage: An instance of ATCStorageBase for data storage.
        Available default classes:
        - ATCMemoryStorage: Stores data in memory.
        - ATCRedisStorage: Stores data in Redis.
    :param manifest_url: URL to the AiogramTonConnect manifest.
    :param redirect_url: URL to the redirect after connecting.
    :param exclude_wallets: List of wallet names to exclude.
    :param qrcode_provider: QRImageProviderBase or QRUrlProviderBase instance.
        Available default classes:
        - QRImageProvider: Generates QR codes locally using the library and
        displays message type as 'photo' with the QR code image.
        - QRUrlProvider: Generates QR codes using a third-party API and
        displays message type as 'text' with the image as 'web_page_preview'.
    :param inline_keyboard: InlineKeyboardBase class for managing inline keyboards.
    """

    def __init__(
            self,
            storage: ATCStorageBase,
            manifest_url: str,
            redirect_url: str = None,
            exclude_wallets: List[str] = None,
            text_message: Optional[Type[TextMessageBase]] = None,
            inline_keyboard: Optional[Type[InlineKeyboardBase]] = None,
            qrcode_provider: Optional[Union[QRImageProviderBase, QRUrlProviderBase]] = None,
    ) -> None:
        self.storage = storage
        self.manifest_url = manifest_url
        self.redirect_url = redirect_url
        self.exclude_wallets = exclude_wallets

        self.qrcode_provider = qrcode_provider or QRImageProvider()
        self.text_message = text_message or TextMessage
        self.inline_keyboard = inline_keyboard or InlineKeyboard

    async def __call__(
            self,
            handler: Callable[[TelegramObject, Dict[str, Any]], Awaitable[Any]],
            event: TelegramObject,
            data: Dict[str, Any],
    ) -> Any:
        """
        Execute middleware

        :param handler: Wrapped handler in middlewares chain
        :param event: Incoming event (Subclass of :class:`aiogram.types.base.TelegramObject`)
        :param data: Contextual data. Will be mapped to handler arguments
        :return: :class:`Any`
        """
        state: FSMContext = data.get("state")
        state_data = await state.get_data()

        account_wallet = state_data.get("account_wallet")
        account_wallet = AccountWallet(**account_wallet) if account_wallet else None
        data["account_wallet"] = account_wallet

        info_wallet = state_data.get("info_wallet")
        info_wallet = InfoWallet(**info_wallet) if info_wallet else None
        data["info_wallet"] = info_wallet

        app_wallet = state_data.get("app_wallet")
        app_wallet = AppWallet(**app_wallet) if app_wallet else None
        data["app_wallet"] = app_wallet

        user: User = data.get("event_from_user")
        language_code = state_data.get("language_code")
        language_code = language_code if language_code else user.language_code
        last_transaction_boc = state_data.get("last_transaction_boc")

        atc_user = ATCUser(
            id=user.id,
            language_code=language_code,
            last_transaction_boc=last_transaction_boc,
            info_wallet=info_wallet,
            app_wallet=app_wallet,
            account_wallet=account_wallet,
        )
        data["atc_user"] = atc_user

        tonconnect = AiogramTonConnect(
            storage=SessionStorage(self.storage, atc_user.id),
            manifest_url=self.manifest_url,
            redirect_url=self.redirect_url,
            exclude_wallets=self.exclude_wallets
        )
        atc_manager = ATCManager(
            storage=self.storage,
            tonconnect=tonconnect,
            text_message=self.text_message(atc_user.language_code),
            inline_keyboard=self.inline_keyboard(atc_user.language_code),
            qrcode_provider=self.qrcode_provider,
            user=atc_user,
            data=data,
        )
        data["atc_manager"] = atc_manager

        return await handler(event, data)



from .callback import ConnectWalletCallbackStorage
from .callback import SendTransactionCallbackStorage
from .session import SessionStorage
from .task import TaskStorage

__all__ = [
    "ConnectWalletCallbackStorage",
    "SendTransactionCallbackStorage",
    "SessionStorage",
    "TaskStorage",
]



from enum import IntEnum
from typing import Awaitable, Callable, Dict, List, Optional, Any, Union

from pydantic import BaseModel, Field
from pytonconnect.parsers import WalletInfo

from aiogram_tonconnect.utils.address import Address


class AppWallet(BaseModel):
    """
    Data class representing a AiogramTonConnect application wallet.

    - app_name: Name of the application.
    - name: Name of the wallet.
    - image: URL of the wallet image.
    - bridge: List of dictionaries representing the bridge.
    - bridge_url: Optional URL of the bridge.
    - platforms: Optional list of supported platforms.
    - tondns: Optional TON DNS information.
    - about_url: Optional URL providing information about the wallet.
    - universal_url: Optional universal URL associated with the wallet.
    """

    app_name: str
    name: str
    image: str
    bridge: List[Dict]
    bridge_url: Optional[str] = Field(default=None)
    platforms: List[str] = Field(default=None)
    tondns: Optional[str] = Field(default=None)
    about_url: Optional[str] = Field(default=None)
    universal_url: Optional[str] = Field(default=None)


class AccountWallet(BaseModel):
    """
    Data class representing a AiogramTonConnect account wallet.

    - address: Optional wallet address.
    - state_init: Optional wallet state initialization.
    - public_key: Optional public key associated with the wallet.
    - chain: Optional chain information.
    """

    address: Optional[Address] = Field(default=None)
    state_init: Optional[str] = Field(default=None)
    public_key: Optional[str] = Field(default=None)
    chain: Optional[int] = Field(default=None)


class DeviceInfo(BaseModel):
    """
    Data class representing a device information.

    - platform: Device platform (e.g. "iphone" | "ipad" | "android").
    - app_name: Application name (e.g. "Tonkeeper").
    - app_version: Application version (e.g. "2.3.367").
    - max_protocol_version: Maximum protocol version supported by the device.
    - features: List of device features.
    """
    platform: str
    app_name: str
    app_version: str
    max_protocol_version: int
    features: List[Union[dict, str]]


class TonProof(BaseModel):
    """
    Data class representing a TON Proof.

    - timestamp: Timestamp of the proof.
    - domain_len: Length of the domain.
    - domain_val: Domain value.
    - payload: Payload of the proof.
    - signature: Signature of the proof.
    """
    timestamp: int
    domain_len: int
    domain_val: str
    payload: str
    signature: str


class InfoWallet(BaseModel):
    """
    Data class representing information about user's wallet.

    - device: Information about user's wallet's device.
    - provider: Provider type.
    - account: Selected account.
    - ton_proof: Response for ton_proof item request.
    """
    device: Optional[DeviceInfo] = Field(default=None)
    provider: Optional[str] = Field(default="http")
    account: Optional[AccountWallet] = Field(default=None)
    ton_proof: Optional[TonProof] = Field(default=None)

    @staticmethod
    def from_pytonconnect_wallet(wallet: WalletInfo) -> 'InfoWallet':
        return InfoWallet(
            device=DeviceInfo(
                platform=wallet.device.platform,
                app_name=wallet.device.app_name,
                app_version=wallet.device.app_version,
                max_protocol_version=wallet.device.max_protocol_version,
                features=wallet.device.features,
            ) if wallet.device else None,
            provider=wallet.provider,
            account=AccountWallet(
                address=Address(hex_address=wallet.account.address),
                state_init=wallet.account.wallet_state_init,
                public_key=wallet.account.public_key,
                chain=wallet.account.chain,
            ) if wallet.account else None,
            ton_proof=TonProof(
                timestamp=wallet.ton_proof.timestamp,
                domain_len=wallet.ton_proof.domain_len,
                domain_val=wallet.ton_proof.domain_val,
                payload=wallet.ton_proof.payload,
                signature=wallet.ton_proof.signature.hex(),
            ) if wallet.ton_proof else None,
        )


class ATCUser(BaseModel):
    """
    Data class representing a AiogramTonConnect user.

    - id: User ID.
    - language_code: Language code associated with the user.
    - app_wallet: Optional AiogramTonConnect application wallet.
    - account_wallet: Optional AiogramTonConnect account wallet.
    - last_transaction_boc: Optional BOC of the last transaction.
    """

    id: int
    language_code: str
    app_wallet: Optional[AppWallet] = Field(default=None)
    info_wallet: Optional[InfoWallet] = Field(default=None)
    account_wallet: Optional[AccountWallet] = Field(default=None)
    last_transaction_boc: Optional[str] = Field(default=None)


class TransactionMessage(BaseModel):
    """
    Data class representing a AiogramTonConnect transaction message.

    - address: Receiver's address.
    - amount: Amount to send in nanoTon.
    - stateInit (string base64, optional): raw once-cell BoC encoded in Base64.
    - payload (string base64, optional): raw one-cell BoC encoded in Base64.

    Example:
        A payload for a TON transfer with commentary:
        payload = urlsafe_b64encode(
            begin_cell()
            .store_uint(0, 32)
            .store_string(comment)
            .end_cell()
            .to_boc()
        ).decode()
    """

    address: str
    amount: str
    stateInit: Optional[str] = Field(default="")
    payload: Optional[str] = Field(default="")


class CHAIN(IntEnum):
    MAINNET = "-239"
    TESTNET = "-3"


class Transaction(BaseModel):
    """
    Data class representing a AiogramTonConnect transaction.

    - valid_until: Validity duration of the transaction.
    - network: The network (mainnet or testnet) where DApp intends to send the transaction.
    If not set, the transaction is sent to the network currently set in the wallet,
    but this is not safe and DApp should always strive to set the network.
    If the network parameter is set, but the wallet has a different network set,
    the wallet should show an alert and DO NOT ALLOW TO SEND this transaction.
    - from_: The sender address in '<wc>:<hex>' format from which DApp intends to send the transaction.
    Current account. address by default.
    - messages: List of transaction messages: min is 1, max is 4.
    """

    valid_until: int
    network: Optional[CHAIN] = Field(default=CHAIN.MAINNET.value)
    from_: Optional[str] = Field(default="", alias="from")
    messages: List[TransactionMessage]

    def model_dump(self, **kwargs) -> dict[str, Any]:
        data = super().model_dump(**kwargs)
        data["from"] = data.pop("from_")
        return data


class ConnectWalletCallbacks(BaseModel):
    """
    Data class representing callbacks for connecting a wallet.

    - before_callback: Callable function to be executed before connecting the wallet.
    - after_callback: Callable function to be executed after connecting the wallet.
    """

    before_callback: Callable[..., Awaitable]
    after_callback: Callable[..., Awaitable]


class SendTransactionCallbacks(BaseModel):
    """
    Data class representing callbacks for sending a transaction.

    - before_callback: Callable function to be executed before sending the transaction.
    - after_callback: Callable function to be executed after sending the transaction.
    """

    before_callback: Callable[..., Awaitable]
    after_callback: Callable[..., Awaitable]



from contextlib import suppress

from aiogram import Dispatcher, Router, F
from aiogram.enums import ChatType
from aiogram.types import CallbackQuery
from pytonconnect.exceptions import WalletNotConnectedError

from .manager import ATCManager
from .utils.states import TcState


class AiogramTonConnectHandlers:

    @staticmethod
    async def connect_wallet_callback_query_handler(
            call: CallbackQuery,
            atc_manager: ATCManager,
    ) -> None:
        """
        Handle callback queries related to connecting a wallet.

        :param call: The CallbackQuery instance.
        :param atc_manager: An instance of ATCManager.
        """
        atc_manager.task_storage.remove()

        if atc_manager.tonconnect.connected:
            with suppress(WalletNotConnectedError):
                await atc_manager.tonconnect.disconnect()

        if call.data.startswith("app_wallet:"):
            wallets = await atc_manager.tonconnect.get_wallets()
            app_wallet_name = call.data.split(":")[1]
            app_wallet = next((w for w in wallets if w.app_name == app_wallet_name), wallets[0])
            await atc_manager.state.update_data(app_wallet=app_wallet.model_dump())
            await atc_manager.retry_connect_wallet()

        elif call.data == "back":
            callbacks = await atc_manager.connect_wallet_callbacks_storage.get()
            await callbacks.before_callback(**atc_manager.middleware_data)

        await call.answer()

    @staticmethod
    async def connect_wallet_proof_wrong_callback_query_handler(
            call: CallbackQuery,
            atc_manager: ATCManager,
    ) -> None:
        """
        Handle callback queries related to wrong proof during wallet connection.

        :param call: The CallbackQuery instance.
        :param atc_manager: An instance of ATCManager.
        """
        if call.data == "retry":
            await atc_manager.retry_connect_wallet()
        elif call.data == "back":
            callbacks = await atc_manager.connect_wallet_callbacks_storage.get()
            await callbacks.before_callback(**atc_manager.middleware_data)

        await call.answer()

    @staticmethod
    async def connect_wallet_timeout_callback_query_handler(
            call: CallbackQuery,
            atc_manager: ATCManager,
    ) -> None:
        """
        Handle callback queries related to connection timeout during wallet connection.

        :param call: The CallbackQuery instance.
        :param atc_manager: An instance of ATCManager.
        """
        if call.data == "retry":
            await atc_manager.retry_connect_wallet()
        elif call.data == "back":
            callbacks = await atc_manager.connect_wallet_callbacks_storage.get()
            await callbacks.before_callback(**atc_manager.middleware_data)

        await call.answer()

    @staticmethod
    async def send_transaction_callback_query_handler(
            call: CallbackQuery,
            atc_manager: ATCManager,
    ) -> None:
        """
        Handle callback queries related to sending a transaction.

        :param call: The CallbackQuery instance.
        :param atc_manager: An instance of ATCManager.
        """
        atc_manager.task_storage.remove()

        if call.data == "back":
            callbacks = await atc_manager.send_transaction_callbacks_storage.get()
            atc_manager.middleware_data["account_wallet"] = atc_manager.user.account_wallet
            await callbacks.before_callback(**atc_manager.middleware_data)

        await call.answer()

    @staticmethod
    async def send_transaction_timeout_callback_query_handler(
            call: CallbackQuery,
            atc_manager: ATCManager,
    ) -> None:
        """
        Handle callback queries related to transaction timeout during sending.

        :param call: The CallbackQuery instance.
        :param atc_manager: An instance of ATCManager.
        """
        if call.data == "retry":
            await atc_manager.retry_last_send_transaction()
        elif call.data == "back":
            callbacks = await atc_manager.send_transaction_callbacks_storage.get()
            atc_manager.middleware_data["account_wallet"] = atc_manager.user.account_wallet
            await callbacks.before_callback(**atc_manager.middleware_data)

        await call.answer()

    @staticmethod
    async def send_transaction_rejected_callback_query_handler(
            call: CallbackQuery,
            atc_manager: ATCManager,
    ) -> None:
        """
        Handle callback queries related to rejected transactions.

        :param call: The CallbackQuery instance.
        :param atc_manager: An instance of ATCManager.
        """
        if call.data == "retry":
            await atc_manager.retry_last_send_transaction()
        elif call.data == "back":
            callbacks = await atc_manager.send_transaction_callbacks_storage.get()
            atc_manager.middleware_data["account_wallet"] = atc_manager.user.account_wallet
            await callbacks.before_callback(**atc_manager.middleware_data)

        await call.answer()

    def register(self, dp: Dispatcher):
        """
        Register AiogramTonConnect-related handlers with the given Dispatcher.

        :param dp: The Dispatcher instance.
        """
        router = Router()
        router.callback_query.filter(F.message.chat.type == ChatType.PRIVATE)

        router.callback_query.register(
            self.connect_wallet_callback_query_handler,
            TcState.connect_wallet,
        )
        router.callback_query.register(
            self.connect_wallet_proof_wrong_callback_query_handler,
            TcState.connect_wallet_proof_wrong,
        )
        router.callback_query.register(
            self.connect_wallet_timeout_callback_query_handler,
            TcState.connect_wallet_timeout,
        )
        router.callback_query.register(
            self.send_transaction_callback_query_handler,
            TcState.send_transaction,
        )
        router.callback_query.register(
            self.send_transaction_timeout_callback_query_handler,
            TcState.send_transaction_timeout,
        )
        router.callback_query.register(
            self.send_transaction_rejected_callback_query_handler,
            TcState.send_transaction_rejected,
        )
        dp.include_router(router)



import time
from base64 import urlsafe_b64encode
from typing import Optional, Union

from pytoniq_core import begin_cell, Cell

from .models import Transaction, TransactionMessage

__all__ = [
    "TONTransferTransaction",
    "JettonTransferTransaction",
    "NFTTransferTransaction",
]


class TONTransferTransaction(Transaction):
    """
    Create a TON (Telegram Open Network) transfer transaction.

    :param address: The address to which the transfer is made.
    :param amount: The amount in TON to be transferred.
    :param comment: An optional comment for the transaction.
    """

    def __init__(
            self,
            address: str,
            amount: Union[int, float],
            comment: Optional[str] = "",
    ) -> None:
        payload = urlsafe_b64encode(
            begin_cell()
            .store_uint(0, 32)
            .store_string(comment)
            .end_cell()
            .to_boc()
        ).decode()
        super().__init__(
            messages=[
                TransactionMessage(
                    address=address,
                    payload=payload,
                    amount=str(int(amount * 1e9)),
                ),
            ],
            valid_until=int(time.time() + 300),
        )


class JettonTransferTransaction(Transaction):
    """
    Create a Jetton transfer transaction.
    Jetton specification link https://github.com/ton-blockchain/TEPs/blob/master/text/0074-jettons-standard.md

    :param jetton_wallet_address: The address of the Jetton wallet.
    :param recipient_address: The recipient's address.
    :param jetton_amount: The amount of Jetton to transfer.
    :param jetton_decimal: The number of decimal places in the Jetton amount (default is 9).
    :param query_id: The query ID (default is 0).
    :param transfer_fee: The transfer fee amount in TON (default is 0.05 TON).
    :param response_address: The address for the response (the sender's address is specified).
        If the address is specified, the excess of TON is returned to the specified address;
        otherwise, it remains on the token contract.
    :param custom_payload: Custom payload for the transaction.
    :param forward_payload: Forward payload for the transaction.
        If forward_amount is greater than 0, this payload will be included with the notification to the new owner.
    :param forward_amount: Forward amount in TON (default is 0.01 TON).
        A notification will be sent to the new owner if the amount is greater than 0;
        otherwise, the new owner will not receive a notification.
    """

    def __init__(
            self,
            jetton_wallet_address: str,
            recipient_address: str,
            jetton_amount: int,
            jetton_decimal: int = 9,
            query_id: Optional[int] = 0,
            transfer_fee: Union[float, int] = 0.05,
            response_address: Optional[str] = None,
            custom_payload: Optional[Cell] = Cell.empty(),
            forward_payload: Optional[Cell] = Cell.empty(),
            forward_amount: Union[float, int] = 0.01,
    ) -> None:
        payload = urlsafe_b64encode(
            begin_cell()
            .store_uint(0xf8a7ea5, 32)
            .store_uint(query_id, 64)
            .store_coins(int(jetton_amount * 10 ** jetton_decimal))
            .store_address(recipient_address)
            .store_address(response_address)
            .store_maybe_ref(custom_payload)
            .store_coins(int(forward_amount * 1e9))
            .store_maybe_ref(forward_payload)
            .end_cell()
            .to_boc()
        ).decode()
        super().__init__(
            messages=[
                TransactionMessage(
                    address=jetton_wallet_address,
                    payload=payload,
                    amount=str(int(transfer_fee * 1e9)),
                ),
            ],
            valid_until=int(time.time() + 300),
        )


class NFTTransferTransaction(Transaction):
    """
    Create an NFT (Non-Fungible Token) transfer transaction.
    Link to NFT specification https://github.com/ton-blockchain/TEPs/blob/master/text/0062-nft-standard.md

    :param nft_address: The address of the NFT.
    :param new_owner_address: The new owner address.
    :param response_address: The address for the response (the sender's address is specified).
        If the address is specified, the excess of TON is returned to the specified address;
        otherwise, it remains on the token contract.
    :param query_id: The query ID (default is 0).
    :param transfer_fee: The transfer fee amount (default is 0.05 TON).
    :param custom_payload: Custom payload for the transaction.
    :param forward_payload: Forward payload for the transaction.
        If forward_amount is greater than 0, this payload will be included with the notification to the new owner.
    :param forward_amount: Forward amount in TON (default is 0.01 TON).
        A notification will be sent to the new owner if the amount is greater than 0;
        otherwise, the new owner will not receive a notification.
    """

    def __init__(
            self,
            nft_address: str,
            new_owner_address: str,
            response_address: Optional[str] = None,
            query_id: Optional[int] = 0,
            transfer_fee: Union[int, float] = 0.05,
            custom_payload: Optional[Cell] = Cell.empty(),
            forward_payload: Optional[Cell] = Cell.empty(),
            forward_amount: Union[float, int] = 0.01,
    ) -> None:
        payload = urlsafe_b64encode(
            begin_cell()
            .store_uint(0x5fcc3d14, 32)
            .store_uint(query_id, 64)
            .store_address(new_owner_address)
            .store_address(response_address)
            .store_maybe_ref(custom_payload)
            .store_coins(int(forward_amount * 1e9))
            .store_maybe_ref(forward_payload)
            .end_cell()
            .to_boc()
        ).decode()
        super().__init__(
            messages=[
                TransactionMessage(
                    address=nft_address,
                    payload=payload,
                    amount=str(int(transfer_fee * 1e9)),
                ),
            ],
            valid_until=int(time.time() + 300),
        )



import hashlib
from datetime import datetime
from nacl.utils import random
from nacl.signing import VerifyKey
from nacl.encoding import HexEncoder

from pytonconnect.logger import _LOGGER  # noqa

from aiogram_tonconnect.tonconnect.models import InfoWallet


def generate_payload(ttl: int = 600) -> str:
    payload = bytearray(random(8))

    ts = int(datetime.now().timestamp()) + ttl
    payload.extend(ts.to_bytes(8, 'big'))

    return payload.hex()


def check_payload(payload: str, info_wallet: InfoWallet) -> bool:
    if len(payload) < 32:
        return False
    if not check_proof(payload, info_wallet):
        return False
    ts = int(payload[16:32], 16)
    if datetime.now().timestamp() > ts:
        return False
    return True


def check_proof(src_payload: str, info_wallet: InfoWallet) -> bool:
    if not info_wallet.ton_proof:
        return False

    wc, whash = info_wallet.account.address.hex_address.split(':', maxsplit=2)

    message = bytearray('ton-proof-item-v2/'.encode())
    message.extend(int(wc, 10).to_bytes(4, 'little'))
    message.extend(bytes.fromhex(whash))
    message.extend(info_wallet.ton_proof.domain_len.to_bytes(4, 'little'))
    message.extend(info_wallet.ton_proof.domain_val.encode())
    message.extend(info_wallet.ton_proof.timestamp.to_bytes(8, 'little'))
    payload_to_use = src_payload.encode() if src_payload is not None else info_wallet.ton_proof.payload.encode()
    message.extend(payload_to_use)

    signature_message = bytearray.fromhex('ffff') + 'ton-connect'.encode() + hashlib.sha256(message).digest()

    try:
        verify_key = VerifyKey(info_wallet.account.public_key, HexEncoder)  # type: ignore
        verify_key.verify(hashlib.sha256(signature_message).digest(), bytes.fromhex(info_wallet.ton_proof.signature))
        return True
    except Exception as e:
        _LOGGER.error(f"Error verifying signature: {e}")

    return False



import base64
import struct

from pydantic import BaseModel

__all__ = [
    "Address",
    "raw_to_userfriendly",
]


class Address(BaseModel):
    """
    Represents a TON address with methods for user-friendly display and conversion.

    :Attributes:
    - hex_address (str): Raw TON address in the format "workchain_id:key".

    :Methods:
    - __str__(): Returns a user-friendly address with urlSafe=true and bounceable=false format.
    - to_raw(): Returns the raw TON address.
    - to_userfriendly(test_only: bool = False): Returns a user-friendly address with optional testnet support.
    """
    hex_address: str

    def __str__(self) -> str:
        """
        Returns a user-friendly address with urlSafe=true and bounceable=false format.
        """
        return raw_to_userfriendly(self.hex_address)

    def to_raw(self) -> str:
        """
        Returns the raw TON address.
        """
        return self.hex_address

    def to_userfriendly(self, test_only: bool = False) -> str:
        """
        Returns a user-friendly address with optional testnet support.
        """
        return raw_to_userfriendly(self.hex_address, test_only)


def _calculate_crc_xmodem(payload: bytes) -> int:
    crc = 0
    poly = 0x1021  # CRC-16 Xmodem polynomial

    for byte in payload:
        crc ^= (byte << 8)  # XOR with the next byte

        for _ in range(8):
            if crc & 0x8000:
                crc = (crc << 1) ^ poly
            else:
                crc <<= 1

    return crc & 0xFFFF  # Keep only the lower 16 bits


def raw_to_userfriendly(address: str, test_only: bool = False) -> str:
    """
    Converts a raw address string to a user-friendly format.

    :param address: The raw address string in the format "workchain_id:key".
    :param test_only: The flag indicating if the address is bounceable. Defaults to False.
    :return: The user-friendly address string, encoded in base64 and URL-safe.
    """
    tag = 0x51
    if test_only:
        tag ^= 0x80
    workchain_id, key = address.split(':')
    workchain_id = int(workchain_id)
    key = bytearray.fromhex(key)

    short_ints = [j * 256 + i for i, j in zip(*[iter(key)] * 2)]
    payload = struct.pack(f'Bb{"H" * 16}', tag, workchain_id, *short_ints)
    crc = _calculate_crc_xmodem(payload)
    encoded_key = payload + struct.pack('>H', crc)

    return base64.urlsafe_b64encode(encoded_key).decode("utf-8")



import json
from pathlib import Path
from typing import Any, Dict, List, Union

import aiofiles
from aiogram.client.session import aiohttp
from cachetools import TTLCache
from pytonconnect.exceptions import FetchWalletsError

from ..models import AppWallet


class _FallbackWalletManager:
    """
    Manager class for handling fallback wallets from a local file.
    """

    BASE_DIR = Path(__file__).parent

    def __init__(self, file_path: Union[str, Path] = None) -> None:
        """
        Initialize the FallbackWalletManager.

        :param file_path: Optional path to the fallback wallets file.
        """
        if file_path is None:
            file_path = self.BASE_DIR / 'wallets_fallback.json'
        self.file_path = file_path

    async def get_wallets(self) -> List[Dict[str, Any]]:
        """
        Retrieve wallets from the fallback file.

        :return: List of dictionaries representing wallets.
        """
        async with aiofiles.open(self.file_path, 'r') as f:
            return json.loads(await f.read())

    async def save_wallets(self, wallets: List[Dict[str, Any]]) -> None:
        """
        Save wallets to the fallback file.

        :param wallets: List of dictionaries representing wallets.
        """
        async with aiofiles.open(self.file_path, 'w') as f:
            await f.write(json.dumps(wallets))


class _CachedWalletManager:
    """
    Manager class for handling cached wallets with a TTL cache.
    """

    def __init__(self, cache_ttl: int = None) -> None:
        """
        Initialize the CachedWalletManager.

        :param cache_ttl: Time-to-live for the cache, in seconds.
        """
        if cache_ttl is None:
            cache_ttl = 86400
        self.cache = TTLCache(maxsize=1, ttl=cache_ttl)

    async def get_wallets(self) -> List[Dict[str, Any]]:
        """
        Retrieve wallets from the cache.

        :return: List of dictionaries representing wallets.
        """
        return self.cache.get("wallets")

    async def save_wallets(self, wallets: List[Dict[str, Any]]) -> None:
        """
        Save wallets to the cache.

        :param wallets: List of dictionaries representing wallets.
        """
        self.cache.setdefault("wallets", wallets)


class WalletManager:
    """
    Manager class for handling AiogramTonConnect wallets.

    :param wallets_list_source_url: URL to fetch the wallets list.
    :param exclude_wallets: List of wallet names to exclude.
    :param cache_ttl: Time-to-live for the cache, in seconds.
    :param file_path: Path to the fallback wallets file.
    """

    def __init__(
            self,
            wallets_list_source_url: str =
            'https://raw.githubusercontent.com/ton-blockchain/wallets-list/main/wallets-v2.json',
            exclude_wallets: List[str] = None,
            cache_ttl: int = None,
            file_path: str = None,
    ) -> None:
        """
        Initialize the WalletManager.

        :param wallets_list_source_url: URL to fetch the wallets list.
        :param exclude_wallets: List of wallet names to exclude.
        :param cache_ttl: Time-to-live for the cache, in seconds.
        :param file_path: Path to the fallback wallets file.
        """
        self.cache_manager = _CachedWalletManager(cache_ttl)
        self.fallback_manager = _FallbackWalletManager(file_path)

        self._exclude_wallets = exclude_wallets
        self._wallets_list_source_url = wallets_list_source_url

    async def __fetch(self) -> List[Dict[str, Any]]:
        """
        Fetch the wallets list from the specified URL.

        :return: List of dictionaries representing wallets.
        """
        async with aiohttp.ClientSession() as session:
            async with session.get(self._wallets_list_source_url) as response:
                if response.status == 200:
                    wallets = await response.json(content_type=response.content_type)
                    if not isinstance(wallets, list):
                        raise FetchWalletsError('Wrong wallets list format, wallets list must be an array.')
                    return wallets
                raise FetchWalletsError('Error fetching wallets')

    def __exclude(self, wallets: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """
        Exclude specified wallets from the list.

        :param wallets: List of dictionaries representing wallets.
        :return: List of dictionaries representing wallets after exclusion.
        """
        if self._exclude_wallets is None:
            return wallets
        return [w for w in wallets if w['app_name'] not in self._exclude_wallets]

    @staticmethod
    def __get_supported(wallets: List[Dict[str, Any]]) -> List[Dict]:
        """
        Filter out wallets that do not support the required features.

        :param wallets: List of dictionaries representing wallets.
        :return: List of dictionaries representing supported wallets.
        """
        supported_wallets = []
        for wallet in wallets:
            try:
                wallet_obj = AppWallet(**wallet)
            except (Exception,):
                continue
            for bridge in wallet_obj.bridge:
                if "sse" in bridge.get("type", "") and "url" in bridge:
                    wallet["bridge_url"] = bridge["url"]
                    supported_wallets.append(wallet)
                    break
        return supported_wallets

    async def __save_wallets(self, wallets: List[Dict[str, Any]]) -> None:
        await self.cache_manager.save_wallets(wallets)
        await self.fallback_manager.save_wallets(wallets)

    def __process_wallets(self, wallets: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        excluded_wallets = self.__exclude(wallets)
        supported_wallets = self.__get_supported(excluded_wallets)
        return supported_wallets

    async def get_wallets(self) -> List[AppWallet]:
        """
        Get the list of AiogramTonConnect wallets.

        :return: List of AiogramTonConnect application wallet instances.
        """
        cached_wallets = await self.cache_manager.get_wallets()

        if cached_wallets:
            supported_wallets = self.__process_wallets(cached_wallets)
            return [AppWallet(**w) for w in supported_wallets]
        try:
            default_wallets = await self.__fetch()
        except (Exception,):
            default_wallets = await self.fallback_manager.get_wallets()

        supported_wallets = self.__process_wallets(default_wallets)
        await self.__save_wallets(default_wallets)
        return [AppWallet(**w) for w in supported_wallets]



from abc import ABCMeta
from typing import Any

from redis.asyncio import Redis


class ATCStorageBase(metaclass=ABCMeta):
    """
    Base class for all storage classes.
    """

    async def get_item(self, key: Any, default_value: str = None) -> Any:
        """
        Get an item from the storage.

        :param key: Key for the item.
        :param default_value: Default value if the item is not found.
        :return: Value of the item.
        """
        raise NotImplementedError

    async def set_item(self, key: Any, value: Any) -> None:
        """
        Set an item in the storage.

        :param key: Key for the item.
        :param value: Value of the item.
        """
        raise NotImplementedError

    async def remove_item(self, key: Any) -> None:
        """
        Remove an item from the storage.

        :param key: Key for the item to be removed.
        """
        raise NotImplementedError


class ATCRedisStorage(ATCStorageBase):

    def __init__(self, redis: Redis) -> None:
        """
        :param redis: Redis instance for asynchronous operations.
        """
        self.redis = redis

    async def set_item(self, key: Any, value: Any) -> None:
        async with self.redis.client() as client:
            await client.set(name=key, value=value)

    async def get_item(self, key: Any, default_value: str = None) -> Any:
        async with self.redis.client() as client:
            value = await client.get(name=key)
            return value if value else default_value

    async def remove_item(self, key: Any) -> None:
        async with self.redis.client() as client:
            await client.delete(key)


class ATCMemoryStorage(ATCStorageBase):
    data = {}

    async def set_item(self, key: Any, value: Any) -> None:
        self.data[key] = value

    async def get_item(self, key: Any, default_value: str = None) -> Any:
        return self.data.get(key) if key in self.data else default_value

    async def remove_item(self, key: Any) -> None:
        self.data.pop(key, None)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="aiogram-tonconnect",
    version="0.13.2",
    author="nessshon",
    description="TON Connect UI for aiogram bots.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(exclude=["example", "docs"]),
    install_requires=[
        "aiofiles",
        "aiogram",
        "aiohttp",
        "cachetools",
        "pillow",
        "pydantic",
        "pytonconnect",
        "pytoniq-core",
        "qrcode-styled",
        "redis",
    ],
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    url="https://github.com/nessshon/aiogram-tonconnect/",
    project_urls={
        "Documentation": "https://nessshon.github.io/aiogram-tonconnect/",
        "Bot example": "https://t.me/aiogramTONConnectBot/",
    },
)



from aiogram_tonconnect.tonconnect.storage.base import ATCStorageBase


class SessionStorage:
    KEY_LAST_EVENT_ID = 'last_event_id'
    KEY_CONNECTION = 'connection'

    def __init__(self, storage: ATCStorageBase, user_id: int) -> None:
        """
        Initialize the SessionStorage.

        :param storage: Storage for storing session data.
        :param user_id: User ID associated with the session storage.
        """
        self.storage = storage
        self.user_id = user_id

    def _get_key(self, key: str) -> str:
        """
        Generate a unique key for the session storage.

        :param key: Original key.
        :return: Unique key combining the user_id and original key.
        """
        return f"{self.user_id}:{key}"

    async def set_item(self, key: str, value: str) -> None:
        await self.storage.set_item(self._get_key(key), value)

    async def get_item(self, key: str, default_value: str = None) -> str:
        value = await self.storage.get_item(self._get_key(key), default_value)
        return value.decode() if isinstance(value, bytes) else value

    async def remove_item(self, key: str) -> None:
        await self.storage.remove_item(self._get_key(key))
